#
# Copyright (C) 2025 Masaryk University
#
# MU-INVENIO-CLI is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
"""Main entry point for the MU Invenio CLI application."""

from mu_invenio_cli.cli_context import CLIContext
from mu_invenio_cli.states.create_state import CreateState
from mu_invenio_cli.states.delete_state import DeleteState
from mu_invenio_cli.states.not_selected_state import NotSelectedState
from mu_invenio_cli.states.selected_state import SelectedState
from mu_invenio_cli.states.configuration_state import ConfigurationState
from mu_invenio_cli.states.state import State
from mu_invenio_cli.states.helper_state import HelperState


def main():
    try:
        run()
    except KeyboardInterrupt:
        print("\nExiting the CLI application.")
        exit(0)
    except Exception as e:
        print(f"An error occurred: {e}")
        exit(1)


def run():
    context = CLIContext()
    state = ConfigurationState(context)
    while True:
        state.handle()
        match context.state:
            case State.CONFIGURATION:
                state = ConfigurationState(context)
            case State.NOT_SELECTED:
                state = NotSelectedState(context)
            case State.SELECTED:
                state = SelectedState(context)
            case State.CREATE:
                state = CreateState(context)
            case State.HELP:
                state = HelperState(context)
            case State.DELETE:
                state = DeleteState(context)
            case _:
                state = ConfigurationState(context)


if __name__ == "__main__":
    main()



#
# Copyright (C) 2025 Masaryk University
#
# MU-INVENIO-CLI is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Base uploader class for the MU Invenio CLI application."""

import abc
import json

import requests


def get_file_entity(entries, file_name):
    for entry in entries:
        if entry.get("key") == file_name:
            return entry
    return None


class BaseUploader:
    def __init__(self, context):
        self.context = context
        self.base_api_model_url = f"{self.context.config['BASE_API_URL']}/{self.context.config['MODEL']}"
        self.headers = {
            "Authorization": f"Bearer {self.context.config['API_TOKEN']}",
            "Content-Type": "application/json"
        }

    @abc.abstractmethod
    def upload(self, file_path):
        pass

    def init_file(self, file_name, init_data=None):
        init_url = f"{self.base_api_model_url}/{self.context.selected_id}/draft/files"
        try:
            response = requests.post(
                init_url,
                json=init_data,
                headers=self.headers,
                verify=False
            )
            if response.status_code != 201:
                return {}
            entries = response.json()["entries"]
            return get_file_entity(entries, file_name)
        except requests.exceptions.RequestException as e:
            print(f"Error initializing file upload: {e}")
            return {}

    def commit_file(self, file_name):
        commit_url = f"{self.base_api_model_url}/{self.context.selected_id}/draft/files/{file_name}/commit"
        try:
            response = requests.post(
                commit_url,
                json={},
                headers=self.headers,
                verify=False
            )
            if response.status_code != 200:
                return False
            return True
        except requests.exceptions.RequestException as e:
            print(f"Error commiting file: {e}")
            return False



#
# Copyright (C) 2025 Masaryk University
#
# MU-INVENIO-CLI is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Unit tests for the MU Invenio CLI application states and context management."""

import unittest
from mu_invenio_cli.cli_context import CLIContext
from mu_invenio_cli.states.configuration_state import ConfigurationState
from mu_invenio_cli.states.not_selected_state import NotSelectedState


class TestCLIContext(unittest.TestCase):
    def test_initial_state(self):
        ctx = CLIContext()
        self.assertEqual(ctx.state.name, "CONFIGURATION")
        self.assertEqual(ctx.last_state.name, "CONFIGURATION")
        self.assertIsInstance(ctx.config, dict)


class TestConfigurationState(unittest.TestCase):
    def setUp(self):
        self.ctx = CLIContext()
        self.state = ConfigurationState(self.ctx)

    def test_is_config_complete_false(self):
        self.ctx.config = {"BASE_API_URL": None, "API_TOKEN": None, "MODEL": None}
        self.assertFalse(self.state.is_config_complete())

    def test_is_config_complete_true(self):
        self.ctx.config = {"BASE_API_URL": "url", "API_TOKEN": "token", "MODEL": "model"}
        self.assertTrue(self.state.is_config_complete())


class TestNotSelectedState(unittest.TestCase):
    def setUp(self):
        self.ctx = CLIContext()
        self.state = NotSelectedState(self.ctx)

    def test_clear_state(self):
        self.ctx.selected_id = "123"
        self.ctx.create_json_body = {"a": 1}
        self.ctx.selected_data = {"b": 2}
        self.state.clear_state()
        self.assertIsNone(self.ctx.selected_id)
        self.assertIsNone(self.ctx.create_json_body)
        self.assertIsNone(self.ctx.selected_data)

if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2025 Masaryk University
#
# MU-INVENIO-CLI is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#

"""State for creating a new draft in the MU Invenio CLI application."""

import json
import os

from mu_invenio_cli.services.record_service import RecordService
from mu_invenio_cli.file_selector import FileSelector
from mu_invenio_cli.states.base_state import BaseState
from mu_invenio_cli.states.state import State


class CreateState(BaseState):
    def handle(self):
        while True:
            print("\nCreate Draft Options:")
            print("  1. Set file containing draft json data")
            if self.context.create_json_body:
                print("  2. Print draft json data")
            print("  3. Create")
            print("  8. Back")
            print("  9. Help")
            print("  0. Exit")

            choice = input("Select option: ").strip()
            if choice == '1':
                # file_path = input("Enter path to JSON file: ").strip()
                file_path = FileSelector().select_file_path()
                if not os.path.isfile(file_path):
                    print("File does not exist.")
                    continue
                try:
                    with open(file_path, 'r') as f:
                        self.context.create_json_body = json.load(f)
                    print("File loaded successfully.")
                except Exception as e:
                    print(f"Failed to load JSON: \n     {e}")
            elif self.context.create_json_body and choice == '2':
                self.print_create_json_body()
            elif choice == '3':
                api_client = RecordService(self.context)
                created = api_client.create_draft()
                if created:
                    self.context.selected_data = created
                    self.context.selected_id = created["id"]
                    self.context.state = State.SELECTED
                    self.context.last_state = State.CREATE
                    break
                else:
                    print("Failed to create draft. Please check issue and try again.")
            elif choice == '8':
                self.context.state = State.NOT_SELECTED
                self.context.last_state = State.CREATE
                break
            elif choice == '9':
                self.context.last_state = State.CREATE
                self.context.state = State.HELP
                break
            elif choice == '0':
                exit(0)
            else:
                print("Invalid option. Please try again.")





#
# Copyright (C) 2025 Masaryk University
#
# MU-INVENIO-CLI is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Not selected state for the MU Invenio CLI application.
This state handles the actions when no draft is selected.
"""

from mu_invenio_cli.services.record_service import RecordService
from mu_invenio_cli.states.state import State


class NotSelectedState:
    def __init__(self, context):
        self.context = context

    def has_access_to_draft(self, record_id):
        return True

    def handle(self):
        self.clear_state()

        print("\nPlease create a draft or select an existing one.")
        while True:
            print("\nOptions:")
            print("  1. Create a new draft")
            print("  2. Set draft by id")
            print("  8. Back to configuration")
            print("  9. Help")
            print("  0. Exit")

            choice = input("Select option: ").strip()
            if choice == '1':
                self.context.state = State.CREATE
                self.context.last_state = State.NOT_SELECTED
                print("Switching to create state...")
                break
            elif choice == '2':
                record_id = input("Enter record id: ").strip()
                api_client = RecordService(self.context)
                draft = api_client.get_draft(record_id)
                if draft:
                    self.context.selected_data = draft
                    self.context.selected_id = record_id
                    self.context.last_state = State.NOT_SELECTED
                    self.context.state = State.SELECTED
                    break
                else:
                    print(f"Draft with id {record_id} not found or you do not have access to it.")
                    continue
            elif choice == '8':
                self.context.last_state = State.NOT_SELECTED
                self.context.state = State.CONFIGURATION
                print("Returning to configuration state.")
                break
            elif choice == '9':
                self.context.last_state = State.NOT_SELECTED
                self.context.state = State.HELP
                break
            elif choice == '0':
                exit(0)
            else:
                print("Invalid option. Please try again.")

    def clear_state(self):
        self.context.selected_id = None
        self.context.create_json_body = None
        self.context.selected_data = None



#
# Copyright (C) 2025 Masaryk University
#
# MU-INVENIO-CLI is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""File uploader implementation for the MU Invenio CLI application."""
import requests

from mu_invenio_cli.uploaders.base_uploader import BaseUploader


class FileUploader(BaseUploader):
    def upload(self, file_path):
        print(f"Uploading file: {file_path}")
        file_name = file_path.split("/")[-1]
        init_data = [{
            "key": file_name
        }]
        init_file = self.init_file(file_name, init_data)
        if not init_file:
            return 0
        print(f"[1/3]: File {file_name} - initialized")
        upload_success = self.upload_content(file_path, file_name)
        if not upload_success:
            return -1
        print(f"[2/3]: File content - uploaded")
        commit_success = self.commit_file(file_name)
        if not commit_success:
            return -1
        print(f"[3/3]: File {file_name} - created")
        return 1

    def upload_content(self, file_path, file_name):
        url = f"{self.base_api_model_url}/{self.context.selected_id}/draft/files/{file_name}/content"
        headers = self.headers
        headers["Content-Type"] = "application/octet-stream"
        try:
            with open(file_path, "rb") as f:
                response = requests.put(url, headers=headers, data=f, verify=False)
            if response.status_code != 200:
                return False
            return True
        except Exception as e:
            print(f"Error uploading file content: {e}")
            return False



#
# Copyright (C) 2025 Masaryk University
#
# MU-INVENIO-CLI is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""State for deleting a draft in the MU Invenio CLI application."""

from mu_invenio_cli.services.record_service import RecordService
from mu_invenio_cli.states.base_state import BaseState
from mu_invenio_cli.states.state import State


class DeleteState(BaseState):
    def handle(self):
        while True:
            print(f"\nDo you want to delete draft with id {self.context.selected_id}:")
            print("  1. Confirm")
            print("  2. Cancel")
            print("  9. Help")
            print("  0. Exit")

            choice = input("Select option: ").strip()
            if choice == '1':
                api_client = RecordService(self.context)
                api_client.delete_draft()
                self.context.state = State.NOT_SELECTED
                self.context.last_state = State.DELETE
                break
            elif choice == '2':
                self.context.state = State.SELECTED
                self.context.last_state = State.DELETE
                break
            elif choice == '9':
                self.context.state = State.HELP
                self.context.last_state = State.DELETE
                break
            elif choice == '0':
                exit(0)
            else:
                print("Invalid option, please try again.")



import requests


class RecordService:
    def __init__(self, context):
        self.context = context
        self.base_api_model_url = f"{self.context.config['BASE_API_URL']}/{self.context.config['MODEL']}"
        self.headers = {
            "Authorization": f"Bearer {self.context.config['API_TOKEN']}",
            "Content-Type": "application/json"
        }

    def create_draft(self):
        create_url = f"{self.base_api_model_url}/"
        try:
            response = requests.post(create_url, json=self.context.create_json_body, headers=self.headers, verify=False)
        except requests.exceptions.RequestException as e:
            print(f"Error creating draft: {e}")
            return None
        if response.status_code == 201:
            print("Draft created successfully.")
        else:
            print(f"Failed to create draft: {response.status_code} - {response.text}")
            return None
        response_json = response.json()
        return response_json

    def get_draft(self, draft_id):
        get_url = f"{self.base_api_model_url}/{draft_id}/draft"
        try:
            response = requests.get(get_url, headers=self.headers, verify=False)
        except requests.exceptions.RequestException as e:
            print(f"Error fetching draft: {e}")
            return None
        if response.status_code == 200:
            print("Draft fetched successfully.")
        else:
            print(f"Failed to fetch draft: {response.status_code} - {response.text}")
            return None
        response_json = response.json()
        return response_json

    def delete_draft(self):
        delete_url = f"{self.base_api_model_url}/{self.context.selected_id}/draft"
        try:
            response = requests.delete(delete_url, headers=self.headers, verify=False)
        except requests.exceptions.RequestException as e:
            print(f"Error deleting draft: {e}")
            return False
        if response.status_code == 204:
            print("Draft deleted successfully.")
            return True
        else:
            print(f"Failed to delete draft: {response.status_code} - {response.text}")
            return False



#
# Copyright (C) 2025 Masaryk University
#
# MU-INVENIO-CLI is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""File uploader implementation for the MU Invenio CLI application."""
import os
import aiohttp
import asyncio

from mu_invenio_cli.uploaders.base_uploader import BaseUploader


def calculate_chunks(file_size, chunk_size):
    return (file_size + chunk_size - 1) // chunk_size


def calculate_chunk_size(file_size):
    min_chunk_size = 5 * 1024 * 1024
    max_chunk_size = 100 * 1024 * 1024
    max_chunks = 10000

    chunk_size = max(min_chunk_size, (file_size + max_chunks - 1) // max_chunks)
    return min(chunk_size, max_chunk_size)


class MultipartUploader(BaseUploader):
    def upload(self, file_path):
        print(f"Uploading file: {file_path}")
        file_name = file_path.split("/")[-1]
        file_size = os.path.getsize(file_path)
        chunk_size = calculate_chunk_size(file_size)
        chunks = calculate_chunks(file_size, chunk_size)

        init_data = [{
            "key": file_name,
            "size": file_size,
            "transfer": {
                "type": "M",
                "parts": chunks,
                "part_size": chunk_size
            }
        }]
        init_file = self.init_file(file_name, init_data)
        if not init_file:
            return 0
        print(f"[1/3]: File {file_name} - initialized")
        parts = init_file.get("links", {}).get("parts", {})
        if not parts or len(parts) != chunks:
            print("Error: Invalid parts information from server.")
            return -1
        upload_success = asyncio.run(
            self.upload_parts_async(file_path, file_name, parts, chunk_size)
        )
        if not upload_success:
            return -1
        print(f"[2/3]: File content - uploaded")
        commit_success = self.commit_file(file_name)
        if not commit_success:
            return -1
        print(f"[3/3]: File {file_name} - created")
        return 1

    async def upload_part(self, session, semaphore, file_path, part_info, chunk_size, max_retries=3):
        part_number = part_info["part"]
        url = part_info["url"]
        offset = (part_number - 1) * chunk_size

        async with semaphore:
            for attempt in range(1, max_retries + 1):
                try:
                    with open(file_path, "rb") as f:
                        f.seek(offset)
                        data = f.read(chunk_size)
                    async with session.put(url, data=data) as resp:
                        if resp.status in (200, 201, 204):
                            print(f"Uploaded part {part_number}")
                            return True
                        else:
                            print(f"Failed part {part_number}, status: {resp.status}")
                except Exception as e:
                    print(f"Error uploading part {part_number}: {e}")
                if attempt == max_retries:
                    print(f"Giving up on part {part_number} after {max_retries} attempts.")
                    return False

    async def upload_parts_async(self, file_path, file_name, parts, chunk_size, max_retries=3):
        semaphore = asyncio.Semaphore(10)
        async with aiohttp.ClientSession() as session:
            tasks = [
                self.upload_part(session, semaphore, file_path, part_info, chunk_size, max_retries)
                for part_info in parts
            ]
            results = await asyncio.gather(*tasks)
        if all(results):
            print("All parts uploaded successfully.")
            return True
        else:
            print("Some parts failed to upload.")
            return False



#
# Copyright (C) 2025 Masaryk University
#
# MU-INVENIO-CLI is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Base abstract state class for the MU Invenio CLI application."""

import json
from abc import ABC, abstractmethod


class BaseState(ABC):
    def __init__(self, context):
        self.context = context

    @abstractmethod
    def handle(self):
        pass

    def print_create_json_body(self):
        print("\n" + json.dumps(self.context.create_json_body, indent=4, ensure_ascii=False))

    def print_selected_data(self):
        print("\n" + json.dumps(self.context.selected_data, indent=4, ensure_ascii=False))



#
# Copyright (C) 2025 Masaryk University
#
# MU-INVENIO-CLI is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""File service for handling file uploads in the MU Invenio CLI application."""

import os

import requests

from mu_invenio_cli.file_selector import FileSelector
from mu_invenio_cli.uploaders.file_uploader import FileUploader
from mu_invenio_cli.uploaders.multipart_uploader import MultipartUploader


class FileService:
    def __init__(self, context):
        self.context = context
        self.base_api_model_url = f"{self.context.config['BASE_API_URL']}/{self.context.config['MODEL']}"
        self.headers = {
            "Authorization": f"Bearer {self.context.config['API_TOKEN']}",
            "Content-Type": "application/json"
        }

    def _upload_file(self, file_path):
        if not os.path.isfile(file_path):
            print(f"File does not exist: {file_path}")
            return

        file_size = os.path.getsize(file_path)
        if file_size <= 100 * 1024 * 1024:  # 100MB threshold
            uploaded = FileUploader(self.context).upload(file_path)
        else:
            uploaded = MultipartUploader(self.context).upload(file_path)
        if uploaded < 0:
            print(f"File upload failed: {file_path}")
            deleted = self.delete_file(file_path)
            if not deleted:
                print(f"Failed to clean up after upload failure for {file_path}. Please check the server.")

    def upload_single_file(self):
        file_path = FileSelector().select_file_path()
        if not file_path:
            print("No file selected.")
            return
        self._upload_file(file_path)

    def upload_multiple_files(self):
        file_paths = FileSelector().select_files()
        if not file_paths:
            print("No files selected.")
            return
        for file_path in file_paths:
            self._upload_file(file_path)

    def delete_file(self, file_path):
        file_name = file_path.split("/")[-1]
        delete_url = f"{self.base_api_model_url}/{self.context.selected_id}/draft/files/{file_name}"
        try:
            response = requests.delete(delete_url, headers=self.headers, verify=False)
            if response.status_code != 204:
                print(f"Error deleting file {file_name}: {response.status_code}")
                return False
            print(f"File {file_name} deleted successfully.")
            return True
        except requests.exceptions.RequestException as e:
            print(f"Error deleting file {file_name}: {e}")
            return False



#
# Copyright (C) 2025 Masaryk University
#
# MU-INVENIO-CLI is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""File selector utility using Tkinter for the MU Invenio CLI application."""

import tkinter
from tkinter import filedialog


class FileSelector:
    def __init__(self):
        self.root = tkinter.Tk()
        self.root.withdraw()

    def select_file_path(self):
        file_path = filedialog.askopenfilename()
        return file_path if file_path else None

    def select_files(self):
        file_paths = filedialog.askopenfilenames()
        return file_paths if file_paths else ()



#
# Copyright (C) 2025 Masaryk University
#
# MU-INVENIO-CLI is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Selected state for the MU Invenio CLI application.
This state handles the actions when the draft is selected.
"""

from .base_state import BaseState
from .state import State
from ..services.file_service import FileService


class SelectedState(BaseState):
    def handle(self):
        while True:
            print("\nSelected Draft Options:")
            print("  1. Print draft data")
            print("  2. Edit draft - not implemented yet")
            print("  3. Delete draft")
            print("  4. Upload files")
            print("  8. Back to select draft")
            print("  9. Help")
            print("  0. Exit")

            choice = input("Select option: ").strip()
            if choice == '1':
                self.print_create_json_body()
            elif choice == '2':
                print("Editing draft is not implemented yet.")
            elif choice == '3':
                self.context.state = State.DELETE
                self.context.last_state = State.SELECTED
                break
            elif choice == '4':
                FileService(self.context).upload_multiple_files()
                break
            elif choice == '9':
                self.context.state = State.HELP
                self.context.last_state = State.SELECTED
                break
            elif choice == '0':
                exit(0)
            else:
                print("Invalid option, please try again.")
